module Main exposing (..)

import Auxiliary exposing (..)
import BooleanLexer exposing (lex)
import Browser exposing (..)
import Browser.Dom exposing (..)
import Browser.Events
import Dict exposing (..)
import Html exposing (..)
import Html.Attributes as HtmlAtt exposing (..)
import Html.Events as Event exposing (..)
import Set exposing (Set)
import SolutionDispatcher exposing (solution)
import Svg as Svg exposing (..)
import Svg.Attributes as SvgAtt exposing (..)
import SvgAuxiliary exposing (viewboxHeight, viewboxLeft, viewboxTop, viewboxWidth)
import Types exposing (..)
import ZoomSvg exposing (..)



{--
 Hier kommt die ganze App rein, wo man den Ausdruck eingeben kann 
 (z.B. umgekehrt polnisch), dann die Parsebäume angezeigt werden 
 (oder eine Fehlermeldung), wo man die Blätter rumklicken kann 
 und auch automatisch die Wahrheitstabelle berechnet wird.

 Eventuell könnte man die Zeile in der Wahrheitstabelle, 
 die der gerade angeklickten Belegung entspricht, 
 farblich markieren.
--}


type Msg
    = InputChanged String
    | DrawCanvasMessage ZoomSvg.Message
    | ValueCycle String
    | Resize Float Float
    | ParserChange (List BoolToken -> Result (List StackElement) Formula)


type alias Model =
    { inputStr : String
    , tokenList : List BoolToken
    , parseResult : Result (List StackElement) Formula
    , parseTrees : List Formula
    , truthTable : Html Msg
    , drawCanvas : ZoomSvg Msg
    , assignments : Dict String Assignment
    , viewportWidth : Float
    , viewportHeight : Float
    , parser : List BoolToken -> Result (List StackElement) Formula
    }


type alias Position =
    { x : Float
    , y : Float
    }


type Assignment
    = T
    | F
    | D


main =
    Browser.element
        { init = init
        , view = view
        , update = update
        , subscriptions = subscriptions
        }


subscriptions : Model -> Sub Msg
subscriptions _ =
    Sub.batch
        [ Browser.Events.onResize (\w h -> Resize (toFloat w) (toFloat h))
        ]


initialModel : Model
initialModel =
    { inputStr = ""
    , parseResult = Result.Err []
    , parseTrees = []
    , tokenList = []
    , truthTable = Html.text "no table generated yet :("
    , drawCanvas = ZoomSvg.makeZoomableSvgCanvas DrawCanvasMessage viewboxLeft viewboxTop viewboxWidth viewboxHeight viewboxWidth viewboxHeight
    , assignments = Dict.empty
    , viewportWidth = 0
    , viewportHeight = 0
    , parser = solution.parse
    }



{--
 we initialize the app by simulating an InputChanged "((x & y) | z)"
--}


init : () -> ( Model, Cmd Msg )
init _ =
    update (InputChanged "((x & y) | (!z))") initialModel


update : Msg -> Model -> ( Model, Cmd Msg )
update msg model =
    case msg of
        InputChanged str ->
            let
                tokenList : List BoolToken
                tokenList =
                    lex str

                res : Result (List StackElement) Formula
                res =
                    model.parser tokenList

                mainFormula : Formula
                mainFormula =
                    extractResult res

                unassignedVariables : List String
                unassignedVariables =
                    getUnassignedVariables mainFormula model.assignments

                table : Html Msg
                table =
                    viewTruthTable (solution.generateTruthTable unassignedVariables mainFormula)
            in
            ( { model
                | inputStr = str
                , tokenList = tokenList
                , parseResult = res
                , parseTrees = extractStack res
                , truthTable = table
              }
            , Cmd.none
            )

        DrawCanvasMessage childMessage ->
            ( { model | drawCanvas = ZoomSvg.update childMessage model.drawCanvas }, Cmd.none )

        ValueCycle str ->
            let
                maybeCycle : Maybe Assignment -> Maybe Assignment
                maybeCycle maybe =
                    case maybe of
                        Just T ->
                            Just F

                        Just F ->
                            Just D

                        Just D ->
                            Just T

                        Nothing ->
                            Nothing

                newAssignments : Dict String Assignment
                newAssignments =
                    if Dict.member str model.assignments then
                        Dict.update str maybeCycle model.assignments

                    else
                        Dict.insert str T model.assignments

                mainFormula : Formula
                mainFormula =
                    extractResult model.parseResult

                unassignedVariables : List String
                unassignedVariables =
                    getUnassignedVariables mainFormula newAssignments

                table : Html Msg
                table =
                    viewTruthTable (solution.generateTruthTable unassignedVariables mainFormula)
            in
            ( { model | assignments = newAssignments, truthTable = table }, Cmd.none )

        Resize width height ->
            ( { model | viewportWidth = width, viewportHeight = height }, Cmd.none )

        ParserChange newParser ->
            let
                res : Result (List StackElement) Formula
                res =
                    newParser (lex model.inputStr)

                extracted : Formula
                extracted =
                    extractResult res

                mapHelp : ( String, Assignment ) -> Maybe String
                mapHelp ( s, ass ) =
                    case ass of
                        D ->
                            Nothing

                        _ ->
                            Just s

                varList : List String
                varList =
                    List.filterMap mapHelp (Dict.toList model.assignments)

                table : Html Msg
                table =
                    viewTruthTable (solution.generateTruthTable varList extracted)
            in
            ( { model | parser = newParser, parseTrees = extractStack res, parseResult = res, truthTable = table }, Cmd.none )


css path =
    Html.node "link" [ HtmlAtt.rel "stylesheet", HtmlAtt.href path ] []


getAllVars : Model -> List String
getAllVars model =
    Set.toList (getVariables (extractResult model.parseResult))


getUn2 : Model -> List String
getUn2 model =
    let
        formula =
            extractResult model.parseResult
    in
    getUnassignedVariables formula model.assignments


getUnassignedVariables : Formula -> Dict String Assignment -> List String
getUnassignedVariables formula assignment =
    let
        isUnassigned : String -> Bool
        isUnassigned s =
            case Dict.get s assignment of
                Just T ->
                    False

                Just F ->
                    False

                _ ->
                    True

        allVariables : List String
        allVariables =
            getVariables formula |> Set.toList

        --MARKER
        varList : List String
        varList =
            -- List.filterMap mapHelp (Dict.toList model.assignments)
            List.filter isUnassigned allVariables
    in
    varList


genStartPositions : Position -> Float -> Int -> List Position
genStartPositions pos oldWidth length =
    let
        spacing =
            oldWidth / toFloat (length + 1)

        newY =
            pos.y + 40

        leftBorder =
            oldWidth / 2

        helper : Int -> List Position -> List Position
        helper _ ls =
            case ls of
                [] ->
                    [ { x = pos.x - leftBorder + spacing, y = newY } ]

                a :: rest ->
                    { a | x = a.x + spacing } :: ls
    in
    if length == 1 then
        [ { pos | x = pos.x, y = newY } ]

    else
        List.foldl helper [] (List.range 1 length)


view : Model -> Html Msg
view model =
    let
        rootPos : Position
        rootPos =
            { x = model.drawCanvas.width / 2, y = model.drawCanvas.height / 4 }

        startPos : List Position
        startPos =
            genStartPositions rootPos 400 (List.length model.parseTrees)
    in
    div [ HtmlAtt.class "container-fluid" ]
        [ css "https://maxcdn.bootstrapcdn.com/bootstrap/3.4.1/css/bootstrap.min.css"
        , css "style.css"
        , h1 [] [ Html.text ("Boolean Formulas Programming Project: " ++ solution.name) ]
        , div [ HtmlAtt.class "row h-12" ]
            [ div [ HtmlAtt.class "col-sm-6" ]
                [ p []
                    [ {--label [] [ Html.text "inverse polish" ]
                    , input [ HtmlAtt.type_ "radio", HtmlAtt.name "parseMode", onClick (ParserChange inversePolishParser) ] []
                    , label [] [ Html.text "infix" ]
                    , input [ HtmlAtt.type_ "radio", HtmlAtt.name "parseMode", onClick (ParserChange infixParser) ] []--}
                      textarea
                        [ Event.onInput InputChanged
                        , HtmlAtt.placeholder "Input Boolean Formular here:"
                        , HtmlAtt.value model.inputStr
                        ]
                        []
                    ]
                , p []
                    [ h3 [] [ Html.text "Lex Tokens:" ]
                    , pre [] [ Html.text (Debug.toString model.tokenList) ]
                    ]
                , p []
                    [ h3 [] [ Html.text "Raw Parsing Results:" ]
                    , pre [] [ Html.text (Debug.toString model.parseResult) ]
                    ]

                {--, p []
                    [ h3 [] [ Html.text "Parsed Boolean Formulas:" ]
                    , pre [] [ Html.text (Debug.toString model.parseTrees) ]
                    ]--}
                , p []
                    [ h3 [] [ Html.text "Generated Truthtable:" ]
                    , div [ HtmlAtt.class "table-container" ]
                        [ model.truthTable
                        ]
                    ]
                ]
            , div [ HtmlAtt.class "col-sm-6" ]
                [ div [ HtmlAtt.style "border" "solid black 1pt" ]
                    [ ZoomSvg.view
                        model.drawCanvas
                        []
                        (List.concat (List.map2 (treeToSvg model.drawCanvas model.assignments 400) startPos model.parseTrees))
                    ]
                ]
            ]
        ]



-- depth should start at


treeToSvg : ZoomSvg Msg -> Dict String Assignment -> Float -> Position -> Formula -> List (Svg Msg)
treeToSvg canvas assigned width rootPos tree =
    let
        resultToColor : Formula -> Bool -> String
        resultToColor res isLeaf =
            case res of
                Const b ->
                    case ( b, isLeaf ) of
                        ( True, True ) ->
                            "green"

                        ( True, False ) ->
                            "rgba(200, 255, 200, 1)"

                        ( False, True ) ->
                            "red"

                        ( False, False ) ->
                            "rgba(255,200,200, 1)"

                _ ->
                    if isLeaf then
                        "grey"

                    else
                        "rgba(220,220,220, 1)"

        properAssignment : Dict String Bool
        properAssignment =
            let
                filter k v =
                    case v of
                        D ->
                            False

                        _ ->
                            True

                map k v =
                    case v of
                        T ->
                            True

                        _ ->
                            False
            in
            Dict.map map (Dict.filter filter assigned)

        line : Position -> Position -> Svg Msg
        line root child =
            Svg.line
                [ x1 (String.fromFloat root.x)
                , y1 (String.fromFloat root.y)
                , x2 (String.fromFloat child.x)
                , y2 (String.fromFloat child.y)
                , stroke "black"
                , SvgAtt.strokeWidth (String.fromFloat (1.5 * canvas.zoom))
                ]
                []

        drawNode : Position -> String -> Formula -> Bool -> List (Svg Msg)
        drawNode pos str res events =
            let
                circleAttributes =
                    if events then
                        -- this is a leaf?
                        [ r (String.fromFloat (10.0 * canvas.zoom))
                        , cx (String.fromFloat pos.x)
                        , cy (String.fromFloat pos.y)
                        , stroke "black"
                        , SvgAtt.strokeWidth (String.fromFloat (1.5 * canvas.zoom))
                        , fill (resultToColor res True)
                        , onClick (ValueCycle str)
                        ]

                    else
                        -- this is an inner node?
                        [ r (String.fromFloat (10.0 * canvas.zoom))
                        , cx (String.fromFloat pos.x)
                        , cy (String.fromFloat pos.y)
                        , stroke "black"
                        , SvgAtt.strokeWidth (String.fromFloat (1.5 * canvas.zoom))

                        --, fill (resultToColor res)
                        , fill (resultToColor res False)
                        ]

                ( xOffset, yOffset ) =
                    if events then
                        ( -1 * canvas.zoom * toFloat (String.length str - 1)
                        , -25.0 * canvas.zoom
                        )
                        -- MARKER

                    else
                        ( 0, -4 * canvas.zoom )
            in
            [ Svg.circle circleAttributes []
            , Svg.line
                []
                []
            , Svg.text_
                [ x (String.fromFloat (pos.x + xOffset))
                , y (String.fromFloat (pos.y - yOffset))
                , fontSize (String.fromFloat (18.0 * canvas.zoom))
                , textAnchor "middle"
                , stroke svgTextColor --, stroke "rgb(107, 196, 255)"
                , stroke svgTextColor --, fill "rgb(107, 196, 255)"
                ]
                [ Svg.text str ]
            ]

        withEvents : Bool -> List (Svg Msg)
        withEvents event =
            drawNode rootPos (boolTreeToString tree) (solution.evaluate tree properAssignment) event
    in
    case tree of
        And child1 child2 ->
            let
                newRoots : List Position
                newRoots =
                    genStartPositions rootPos width 2

                newWidth =
                    width / toFloat 2
            in
            List.concat (List.append [ List.map (line rootPos) newRoots ++ withEvents False ] (List.map2 (treeToSvg canvas assigned newWidth) newRoots [ child2, child1 ]))

        Or child1 child2 ->
            let
                newRoots : List Position
                newRoots =
                    genStartPositions rootPos width 2

                newWidth =
                    width / toFloat 2
            in
            List.concat (List.append [ List.map (line rootPos) newRoots ++ withEvents False ] (List.map2 (treeToSvg canvas assigned newWidth) newRoots [ child2, child1 ]))

        Not child1 ->
            let
                newRoots : List Position
                newRoots =
                    genStartPositions rootPos width 1

                newWidth =
                    width / toFloat 1
            in
            List.concat (List.append [ List.map (line rootPos) newRoots ++ withEvents False ] (List.map2 (treeToSvg canvas assigned newWidth) newRoots [ child1 ]))

        Var _ ->
            withEvents True

        Const _ ->
            withEvents False


viewTruthTable : List ( List ( String, Bool ), Formula ) -> Html msg
viewTruthTable table =
    let
        inputOfRow : ( List ( String, Bool ), Formula ) -> List ( String, Bool )
        inputOfRow ( input, _ ) =
            input

        heading : List ( String, Bool ) -> Html msg
        heading input =
            Html.tr [] (List.map (\f -> Html.th [] [ Html.text (Tuple.first f) ]) input ++ [ Html.th [] [ Html.text "Result" ] ])

        rowToHtml : ( List ( String, Bool ), Formula ) -> Html msg
        rowToHtml ( input, res ) =
            let
                permutation : List Bool
                permutation =
                    List.map Tuple.second input
            in
            Html.tr [] (List.map (\f -> Html.td [] [ Html.text (boolToString f) ]) permutation ++ [ Html.td [] [ Html.text (solution.toString res) ] ])

        {--case input of
                [] ->
                    Html.tr [] [ Html.div [] [ Html.text "All variables are disabled (Probably)" ] ]

                _ ->
                    let
                        permutation : List Bool
                        permutation =
                            List.map Tuple.second input
                    in
                    Html.tr [] (List.map (\f -> Html.td [] [ Html.text (boolToString f) ]) permutation ++ [ Html.td [] [ Html.text (solution.toString res) ] ])
                    --}
    in
    case table of
        [] ->
            Html.text "table has no rows"

        a :: _ ->
            Html.table [ HtmlAtt.class "table table-striped" ]
                [ Html.thead []
                    [ heading (inputOfRow a)
                    ]
                , Html.tbody []
                    (List.map rowToHtml table)
                ]


svgTextColor =
    "rgb(70, 140, 190)"
